import photonvision
from robotpy_toolkit_7407.sensors.photonvision.photon_target import PhotonTarget
from wpimath.geometry import Pose3d


class PhotonCamera:
    def __init__(self, name: str, poseRelativeToRobot: Pose3d, scale_constant: float = 1, height=None, pitch=None):
        self.camera = photonvision.PhotonCamera(cameraName=name)
        self.latest_best_target: PhotonTarget = None
        self.latest_targets_all: list[PhotonTarget] = None
        self.camera_to_robot_pose = poseRelativeToRobot
        self.height = height
        self.pitch = pitch
        self.scale_constant = scale_constant

    def hasTargets(self):
        return self.camera.hasTargets()

    def get_scaled_relative_pose(self):
        if self.latest_best_target is not None:
            return self.latest_best_target.relative_pose.translation().toTranslation2d() * self.scale_constant

    def refresh(self):  # Call at the beginning of every loop. Saves target for the loop to optimize lookup times.
        self.latest_targets_all = [PhotonTarget(target) for target in
                                   self.camera.getLatestResult().getTargets()] if self.hasTargets() else None
        self.latest_best_target = PhotonTarget(
            self.camera.getLatestResult().getBestTarget()) if self.hasTargets() else None



from robotpy_toolkit_7407.sensors.gyro.ADIS16448 import GyroADIS16448
from robotpy_toolkit_7407.sensors.gyro.PigeonIMU import PigeonIMUGyro_Wrapper
from robotpy_toolkit_7407.sensors.gyro.base_gyro import Gyro


from wpimath.geometry import Pose3d


class VisionEstimator:
    """
    An estimator (e.g. limelight, photon-vision) that returns a list of robot poses relative to the field.
    """
    def __init__(self):
        pass

    def get_estimated_robot_pose(self) -> list[Pose3d] | None:
        """
        Returns the robot's pose relative to the field, estimated by the vision system. Override this method.
        :return: Vision system estimate of robot pose.
        :rtype: Pose3d | None
        """
        raise NotImplementedError



import math
from wpilib import ADIS16448_IMU

from robotpy_toolkit_7407.utils.units import radians

from robotpy_toolkit_7407.sensors.gyro.base_gyro import Gyro


class GyroADIS16448(Gyro):
    """
    Wrapper class for the ADIS16448 Gyro
    """
    def __init__(self) -> None:
        self._gyro = ADIS16448_IMU()
        self.__offset = 0

    def init(self):
        """
        Initialize the gyro
        """
        self.reset_angle()
        self.__offset = 0

    def get_robot_heading(self) -> radians:
        """
        Returns the angle of the robot's heading in radians (yaw)
        :return: Robot heading (radians)
        """
        return math.radians(self._gyro.getGyroAngleZ() + self.__offset)

    def reset_angle(self):
        """
        Resets the gyro's yaw.
        """
        self.__offset = self.get_robot_heading()



from robotpy_toolkit_7407.oi.buttons import DefaultButton, _Button, AxisButton
from robotpy_toolkit_7407.oi.controllermap import LogitechController, XBoxController
from robotpy_toolkit_7407.oi.joysticks import _Joysticks, JoystickAxis



class Color:
    PURPLE = "\033[95m"
    CYAN = "\033[96m"
    DARKCYAN = "\033[36m"
    BLUE = "\033[94m"
    GREEN = "\033[92m"
    YELLOW = "\033[93m"
    RED = "\033[91m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"
    END = "\033[0m"


class NoColor:
    PURPLE = ""
    CYAN = ""
    DARKCYAN = ""
    BLUE = ""
    GREEN = ""
    YELLOW = ""
    RED = ""
    BOLD = ""
    UNDERLINE = ""
    END = ""



__version__ = '0.5.11'

from robotpy_toolkit_7407.subsystem import Subsystem
from robotpy_toolkit_7407.command import BasicCommand, SubsystemCommand
from robotpy_toolkit_7407.motor import Motor, PIDMotor, EncoderMotor
import robotpy_toolkit_7407.unum



import ctre
import math

from robotpy_toolkit_7407.utils.units import radians
from robotpy_toolkit_7407.sensors.gyro.base_gyro import Gyro


class PigeonIMUGyro_Wrapper(Gyro):
    """
    Wrapper class for the Pigeon2 IMU gyro.
    """

    def __init__(self, port):
        """
        Args:
            port (int): CAN ID of the Pigeon gyro
        """
        self._gyro = ctre.Pigeon2(port)
        self._gyro.configMountPose(0, 0, 0)

    def init(self):
        """
        Initialize gyro
        """
        self.reset_angle()

    def get_robot_heading(self) -> radians:
        """
        Returns the angle of the robot's heading in radians (yaw)
        :return: Robot heading (radians)
        """
        return math.radians(self._gyro.getYaw())

    # reset the gyro
    def reset_angle(self):
        """
        Resets the gyro's yaw.
        """
        self._gyro.setYaw(0)



import commands2


class SubsystemBase(commands2.SubsystemBase):
    """
    All subsystems should inherit from this class for correct usage in commands.
    """
    def __init__(self):
        super().__init__()



from robotpy_toolkit_7407.utils.units import radians, radians_per_second


class Motor:
    def init(self): ...

    def set_raw_output(self, x: float): ...


class EncoderMotor(Motor):
    def get_sensor_position(self) -> radians: ...

    def get_sensor_velocity(self) -> radians_per_second: ...

    def set_sensor_position(self, pos: radians): ...


class PIDMotor(EncoderMotor):
    def set_target_position(self, pos: radians): ...

    def set_target_velocity(self, vel: radians_per_second): ...



from typing import Generic, TypeVar
import commands2

from robotpy_toolkit_7407.subsystem import Subsystem

T = TypeVar("T", bound=Subsystem)


class BasicCommand(commands2.CommandBase):
    """
    Extendable basic command
    """
    ...


class SubsystemCommand(BasicCommand, Generic[T]):
    """
    Extendable subsystem command
    """

    def __init__(self, subsystem: T):
        super().__init__()
        self.subsystem = subsystem
        self.addRequirements(subsystem)



"""Main Unum module.

# TODO: consider alternatives to unum and see how they compare.
"""

class ShouldBeUnitlessError(TypeError):
    """An operation on a Unum failed because it had units unexpectedly."""
    def __init__(self, u):
        TypeError.__init__(self, "expected unitless, got %s" % u)


class IncompatibleUnitsError(TypeError):
    """An operation on two Unums failed because the units were incompatible."""
    def __init__(self, unit1, unit2):
        TypeError.__init__(self, "%s can't be converted to %s" %
                           (unit1.strUnit(), unit2.strUnit()))

class UnumError(Exception):
    """A Unum error occurred that was unrelated to dimensional errors."""
    pass


class ConversionError(UnumError):
    """Failed to convert a unit to the desired type."""
    def __init__(self, u):
        UnumError.__init__(self, "%s has no conversion" % u)


class NameConflictError(UnumError):
    """Tried to define a symbol that was already defined."""
    def __init__(self, unit_key):
        UnumError.__init__(self, "%s is already defined." % unit_key)
        
        
class NonBasicUnitError(UnumError):
    """Expected a basic unit but got a non-basic unit."""
    def __init__(self, u):
        UnumError.__init__(self, "%s not a basic unit" % u)

# With current versions of numpy, we have the following undesirable behavior:
#     >>> array([5,6,7,8]) * M
#     array([5 [m], 6 [m], 7 [m], 8 [m]], dtype=object)
# It seems like array.__div__ is called rather than Unum.__rdiv__,
# with the result that the M is broadcast across the array.
try:
    from numpy import array
    def uarray(array_like, *args, **kwargs):
        """Convenience function to return a Unum containing a numpy array."""
        return Unum.coerceToUnum(array(array_like, *args, **kwargs))
except ImportError:
    pass


class Unum(object):
    """Encapsulates a value attached to a unit.
    
    Implements arithmetic operators, dynamic unit consistency checking, and
    string representation.
    """
    #
    UNIT_SEP = "."
    """Separator between multiple units: e.g. "5 N.m". """
    
    UNIT_DIV_SEP = "/"
    """Separator between numerator and denominator.
    
    If set to None, negative exponents are used instead.
    """
    
    UNIT_FORMAT = "[%s]"
    """Format string for attached unit."""
    
    UNIT_INDENT = " "
    """Separator between value and unit."""
    
    UNIT_HIDE_EMPTY = False
    """If True, unitless unums are displayed as raw numbers."""
    
    UNIT_SORTING = True
    """If True, units are sorted alphabetically for display."""
    
    VALUE_FORMAT = "%s"
    """Format string for value."""
    
    AUTO_NORM = True
    """If True, normalize unums for their string representation."""
        
    # -- internal constants ------------------------------------------
    _NO_UNIT = {}
  
    # -- internal working storage ------------------------------------
    # unit dictionary :
    #  the key is the unit string
    #  the value is a tuple (conversion unum, level, name)
    _unitTable = {}

    __slots__ = ('_value', '_unit', '_normal')

    # TODO: conv is a terrible name throughout. Find replacement?
    def __init__(self, unit, value=1, conv=None, name=''):
        """Create a new unum object.
        
        unit  is a dictionary of {unit symbol : exponent}
        value is a number
        conv  is None if self does not represent a unit (default)
                or 0 if self represents a basic unit
                or a unum equivalent to self, expressed in other unit(s)
                   if self represents a derived unit
        name  is the unit full name if self represents a basic unit
        raises UnumError exception if conv is a unum
                although unit and value do not represent a basic unit
        """
        object.__init__(self)        
        self._value = value
        self._unit = unit
        if conv is None:
            self._normal = False
        else:
            unit_key = list(unit.keys())[0]
            if unit_key in Unum._unitTable:
                raise NameConflictError(unit_key)
            self._normal = True
            if isinstance(conv, int) and conv == 0:
                conv_unum = None
                level = 0
            else:
                if value == 0 or len(unit) != 1 or list(unit.values())[0] != 1:
                    raise NonBasicUnitError(self)
                conv_unum = Unum.coerceToUnum(conv) / value
                level = conv_unum.maxLevel() + 1
                conv_unum._normal = True
            Unum._unitTable[unit_key] = conv_unum, level, name

    def unit(cls, symbol, conv=0, name=''):
        """Return a new unit represented by the string symbol.
        
        If conv is 0, the new unit is a base unit.
        If conv is a Unum, the new unit is a derived unit equal to conv.
        
        >>> KB = Unum.defineUnit("kB", 0, "kilobyte")
        >>> MB = Unum.defineUnit("MB", 1000*KB, "megabyte")
        """
        return cls({symbol:1}, 1, conv, name)
    unit = classmethod(unit)
    
    def reset(cls, unitTable=None):
        """Clear the unit table, replacing it with the new one if provided.
        
        This is generally only useful when playing around with defining new
        units in the interpreter.
        """
        if unitTable is None:
            cls._unitTable = {}
        else:
            cls._unitTable = unitTable
    reset = classmethod(reset)

    def getUnitTable(cls):
        """Return a copy of the unit table."""
        return cls._unitTable.copy()
    getUnitTable = classmethod(getUnitTable)
    
    def copy(self, normalized=False):
        """Return a copy of this Unum, normalizing the copy if specified."""
        result = Unum(self._unit.copy(), self._value)
        if normalized:
            result.normalize()
        return result
        
    def asUnit(self, other):
        """Return a Unum with this Unum's value and the units of the given Unum.
        
        Raises IncompatibleUnitsError if self can't be converted to other.
        Raises NonBasicUnitError if other isn't a basic unit.
        """
        other = Unum.coerceToUnum(other)
        if (other._value == 0) or (other != Unum(other._unit, 1)):
            raise NonBasicUnitError(other)
        s, o = self.matchUnits(other)
        res = Unum(other._unit, s._value / o._value)
        res._normal = True
        return res
    
    def replaced(self, u, conv_unum):
        """Return a Unum with the string u replaced by the Unum conv_unum. 
        
        If u is absent from self, a copy of self is returned.
        """
        res = self.copy() * conv_unum ** self._unit[u]
        del res._unit[u]
        return res

    # Persistence methods (required by __slots__).
    def __getstate__(self):
        return (self._value, self._unit, self._normal)

    def __setstate__(self, state):
        self._value, self._unit, self._normal = state

    # Normalization methods.
    def normalize(self, forDisplay=False):
        """Normalize our units IN PLACE and return self.
        
        Substitutions may be applied to reduce the number of different units,
        while making the fewest substitutions.
        
        If forDisplay is True, then prefer a single unit to no unit.
        # TODO: example of forDisplay.
        # TODO: simplify normalize so it fits in 80 columns...
        """
        best_l = len(self._unit)
        new_subst_unums = [({}, +self)]
        while new_subst_unums:
                subst_unums, new_subst_unums = new_subst_unums, []
                for subst_dict, subst_unum in subst_unums:
                    for u, exp in list(subst_unum._unit.items()):
                        conv_unum = Unum._unitTable[u][0]
                        if conv_unum is not None:
                            new_subst_dict = subst_dict.copy()
                            new_subst_dict[u] = exp + new_subst_dict.get(u, 0)
                            is_new = True
                            for subst_dict2, subst_unum2 in new_subst_unums:
                                if new_subst_dict == subst_dict2:
                                    is_new = False
                                    break
                            if is_new:       
                                s = subst_unum.replaced(u, conv_unum)
                                new_subst_unums.append((new_subst_dict, s))
                                new_l = len(s._unit)
                                if new_l < best_l and not (forDisplay and new_l == 0 and best_l == 1):
                                    self._value, self._unit = s._value, s._unit
                                    best_l = new_l
        return self                       

    def checkNoUnit(self):
        """Raise ShouldBeUnitlessError if self has a unit."""
        if self._unit:
            raise ShouldBeUnitlessError(self)
    
    def maxLevel(self):
        """ returns the maximum level of self's units
        """
        return max([0] + [Unum._unitTable[u][1] for u in self._unit.keys()])

    def matchUnits(self, other):
        """Return (self, other) where both Unums have the same units.
        
        Raises IncompatibleUnitsError if there is no way to do this.
        If there are multiple ways to do this, the units of self, then other 
        are preferred, and then by maximum level.
        """   
        if self._unit == other._unit:
            return self, other
        
        s = self.copy()
        o = other.copy()
        s_length, o_length = len(s._unit), len(o._unit)
        revert = (s_length > o_length or
                 (s_length == o_length and s.maxLevel() < o.maxLevel()))
        if revert:
            s, o = o, s
        target_unum = Unum(s._unit, 1)
        o /= target_unum
        o.normalize()
        if o._unit:
            raise IncompatibleUnitsError(self, other)
        o._unit = s._unit
        if revert:
            s, o = o, s
        return s, o
    
    # TODO: could support in-place operators for 2.5 and higher.
    
    # Arithmetic operations.
    # These raise IncompatibleUnitsError if the operands have incompatible units.
    def __add__(self, other):
        s, o = self.matchUnits(Unum.coerceToUnum(other))
        return Unum(s._unit, s._value + o._value)
    
    def __sub__(self, other):
        s, o = self.matchUnits(Unum.coerceToUnum(other))
        return Unum(s._unit, s._value - o._value)
                    
    def __pos__(self):
        # TODO: is it really beneficial to share the unit dictionary?
        return Unum(self._unit.copy(), self._value)

    def __neg__(self):
        return Unum(self._unit.copy(), -self._value)

    def __mul__(self, other):
        other = Unum.coerceToUnum(other)
        if not self._unit:
            unit = other._unit
        elif not other._unit:
            unit = self._unit
        else:
            unit = self._unit.copy()
            for u, exp in other._unit.items():          
                exp += unit.get(u, 0)
                if exp:
                    unit[u] = exp
                else:
                    del unit[u]
        return Unum(unit, self._value * other._value)

    def __div__(self, other):
        other = Unum.coerceToUnum(other)
        if not other._unit:
            unit = self._unit
        else: 
            unit = self._unit.copy()
            for u, exp in list(other._unit.items()):          
                exp -= unit.get(u, 0)
                if exp:
                    unit[u] = -exp
                else:
                    del unit[u]
        return Unum(unit, self._value / other._value)    
    __truediv__ = __div__ # Python 3.0 compatibility.
    
    def __floordiv__(self, other):
        other = Unum.coerceToUnum(other)
        if not other._unit:
            unit = self._unit
        else: 
            unit = self._unit.copy()
            for u, exp in list(other._unit.items()):          
                exp -= unit.get(u, 0)
                if exp:
                    unit[u] = -exp
                else:
                    del unit[u]
        return Unum(unit, self._value // other._value)         
    
    def __pow__(self, other):
        other = Unum.coerceToUnum(other)
        if other._value:
            other = other.copy(True)
            other.checkNoUnit()       
            unit = self._unit.copy()
            for u in list(self._unit.keys()):
                unit[u] *= other._value
        else:
            unit = Unum._NO_UNIT
        return Unum(unit, self._value ** other._value)

    def __lt__(self, other):
        s, o = self.matchUnits(Unum.coerceToUnum(other))
        return s._value < o._value

    def __le__(self, other):
        s, o = self.matchUnits(Unum.coerceToUnum(other))
        return s._value <= o._value

    def __gt__(self, other):
        s, o = self.matchUnits(Unum.coerceToUnum(other))
        return s._value > o._value

    def __ge__(self, other):
        s, o = self.matchUnits(Unum.coerceToUnum(other))
        return s._value >= o._value

    def __eq__(self, other):
        s, o = self.matchUnits(Unum.coerceToUnum(other))
        return s._value == o._value

    def __ne__(self, other):
        s, o = self.matchUnits(Unum.coerceToUnum(other))
        return s._value != o._value
    
    def __abs__(self):
        return Unum(self._unit.copy(), abs(self._value)) 

    def asNumber(self, other=None):
        """Return the (normalized) raw value of self.
        
        If other is supplied, first convert to other's units before returning
        the raw value.
        
        Raises NonBasicUnitError if other is supplied, but has a value other
        than 1. (e.g., kg.asNumber(2*g) is an error, but kg.asNumber(g) is ok.)            
        """
        if other is None:
            return self.copy(True)._value
        
        if isinstance(other, Unum):
            if (other._value == 0) or (other != Unum(other._unit, 1)):
                raise NonBasicUnitError(other)
            else:
                s, o = self.matchUnits(other)
                return s._value / o._value
        else:
            s = self.copy(True)
            s.checkNoUnit()
            return s._value / other
        
    def __complex__(self):
        return complex(self.asNumber(1))

    def __int__(self):       
        return int(self.asNumber(1))

    def __long__(self):      
        return int(self.asNumber(1))
    
    def __float__(self):       
        return float(self.asNumber(1))

    def __radd__(self, other):
        return Unum.coerceToUnum(other).__add__(self)

    def __rsub__(self, other):         
        return Unum.coerceToUnum(other).__sub__(self)

    def __rmul__(self, other):     
        return Unum.coerceToUnum(other).__mul__(self)

    def __rdiv__(self, other):   
        return Unum.coerceToUnum(other).__div__(self)
    __rtruediv__ = __rdiv__ # Python 3.0 compatibility.

    def __rfloordiv__(self, other):
        return Unum.coerceToUnum(other).__floordiv__(self)
        
    def __rpow__(self, other):         
        return Unum.coerceToUnum(other).__pow__(self)

    def __getitem__(self, index):
        return Unum(self._unit, self._value[index])

    def __setitem__(self, index, value):
        u = Unum.coerceToUnum(value)        
        self._value[index] = u.asNumber(Unum(self._unit, 1))

    def __len__(self):
        return len(self._value)

    # -- String representation methods -------------------------------
    def strUnit(self):
        """Return a string representation of our unit."""
        def fmt(exp):
            f = ''
            if exp != 1:
                f = str(exp)
            return f
        numer, denom = '', ''
        units = list(self._unit.items())       
        if Unum.UNIT_SORTING:
            units.sort()
        for u, exp in units:          
            if exp > 0 or not Unum.UNIT_DIV_SEP:
                if numer:
                    numer += Unum.UNIT_SEP
                numer += u + fmt(exp)
            else:
                if denom:
                    denom += Unum.UNIT_SEP
                denom += u + fmt(-exp)
        if denom:
            denom = Unum.UNIT_DIV_SEP + denom
            if not numer:
                numer = '1'
        if not numer and Unum.UNIT_HIDE_EMPTY:
            result = ''
        else:
            result = Unum.UNIT_FORMAT % (numer + denom)  
        return result

    def __str__(self):
        """Return our string representation, normalized if applicable.
        
        Normalization occurs if Unum.AUTO_NORM is set.
        """
        if Unum.AUTO_NORM and not self._normal:
            self.normalize(True)
            self._normal = True  
        return (Unum.VALUE_FORMAT % self._value + 
                Unum.UNIT_INDENT + 
                self.strUnit())
    __repr__ = __str__

    # TODO: what is converted method for?
    def converted(self): 
        """ returns self converted following _unitTable
            raises UnumError exception if the self's unit is not unique
             or if no conversion exists for self
        """
        def fix(u):
            """Prevent implicit normalization of self."""
            u._normal = True
            return u
        
        self._normal = True
        return self
        if len(self._unit) != 1:
            raise ConversionError(fix(+self))
        u = list(self._unit.keys())[0]
        conv = Unum._unitTable[u][0]
        if conv is None:
            raise ConversionError(self)    
        return fix(self.replaced(u, conv))

    def coerceToUnum(value):
        """Return a unitless Unum if value is a number.
        
        If value is a Unum already, it is returned unmodified.
        """
        if isinstance(value, Unum):
            return value
        else:
            return Unum(Unum._NO_UNIT, value)
    coerceToUnum = staticmethod(coerceToUnum)



import wpilib


class LimitSwitch:
    """
    Wrapper class for I2C Limit Switches
    """

    def __init__(self, port: int, inverted: bool = True):
        """Wrapper class for I2C Limit Switches
        Args:
            port (int): I2C port of the limit switch.
            inverted (bool, optional): Return the inverted boolean of output. Defaults to True.
        """
        self.limit_switch = wpilib.DigitalInput(port)
        self.reverse = inverted

    def get_value(self):
        """Return if the limit switch is pressed or if object is detected (in the case of non-tactile sensors).

        Returns:
            bool: True if pressed, False if not.
        """
        if self.reverse:
            return not self.limit_switch.get()
        return self.limit_switch.get()


class MagneticLimitSwitch(LimitSwitch):
    """
    Wrapper class for I2C Magnetic Limit Switches
    """

    def __init__(self, port: int, inverted: bool = True):
        """Wpilib I2C Magnetic Limit Switch Wrapper extends LimitSwitch

        Args:
            port (int): I2C port of the limit switch.
            inverted (bool, optional): Return the inverted boolean of output. Defaults to True.
        """
        super().__init__(port, inverted)


class PhotoElectricSwitch(LimitSwitch):
    """
    Wrapper class for I2C PhotoElectric Switches
    """

    def __init__(self, port: int, inverted: bool = True):
        """Wpilib I2C PhotoElectric Limit Switch Wrapper extends LimitSwitch

        Args:
            port (int): I2C port of the limit switch.
            inverted (bool, optional): Return the inverted boolean of output. Defaults to True.
        """
        super().__init__(port, inverted)



from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

import ctre
from robotpy_toolkit_7407.unum import Unum

from robotpy_toolkit_7407.motor import PIDMotor
from robotpy_toolkit_7407.utils.units import rad, rev, s, radians_per_second, radians_per_second_squared, radians


@dataclass
class TalonConfig:
    """
    Configuration for a TalonSRX/Falcon motor controller

    Args:
        kP: Proportional gain
        kI: Integral gain
        kD: Derivative gain
        kF: Feedforward gain
        closed_loop_peak_output: The maximum output of the controller
        neutral_brake: Whether to brake or coast when the motor is not moving
    """
    k_P: Optional[float] = None
    k_I: Optional[float] = None
    k_D: Optional[float] = None
    k_F: Optional[float] = None
    closed_loop_peak_output: Optional[float] = None
    motion_cruise_velocity: Optional[radians_per_second] = None
    motion_acceleration: Optional[radians_per_second_squared] = None
    neutral_brake: Optional[bool] = None
    integral_zone: Optional[float] = None
    max_integral_accumulator: Optional[float] = None


talon_sensor_unit = Unum.unit("talon_sensor_u", rev / 2048, "talon sensor unit")
hundred_ms = Unum.unit("100ms", s / 10, "100 milliseconds")
talon_sensor_vel_unit = talon_sensor_unit / hundred_ms
talon_sensor_accel_unit = talon_sensor_vel_unit / s

k_sensor_pos_to_radians = talon_sensor_unit.asNumber(rad)
k_radians_to_sensor_pos = rad.asNumber(talon_sensor_unit)
k_sensor_vel_to_rad_per_sec = talon_sensor_vel_unit.asNumber(rad / s)
k_rad_per_sec_to_sensor_vel = (rad / s).asNumber(talon_sensor_unit / hundred_ms)
k_sensor_accel_to_rad_per_sec_sq = talon_sensor_accel_unit.asNumber(rad / (s * s))
k_rad_per_sec_sq_to_sensor_accel = (rad / (s * s)).asNumber(talon_sensor_unit / (s * hundred_ms))


class _Talon(PIDMotor):
    _motor: ctre.BaseTalon

    def __init__(self, can_id: int, inverted: bool = False, config: TalonConfig = None):
        super().__init__()
        self._can_id = can_id
        self._config = config
        self._inverted = inverted

    def get_sensor_position(self) -> radians:
        return self._motor.getSelectedSensorPosition(0) * k_sensor_pos_to_radians

    def set_sensor_position(self, pos: radians):
        self._motor.setSelectedSensorPosition(pos * k_radians_to_sensor_pos)

    def get_sensor_velocity(self) -> radians_per_second:
        return self._motor.getSelectedSensorVelocity(0) * k_sensor_vel_to_rad_per_sec

    def set_raw_output(self, x: float):
        self._motor.set(ctre.ControlMode.PercentOutput, x)

    def set_target_position(self, pos: radians):
        self._motor.set(ctre.ControlMode.MotionMagic, pos * k_radians_to_sensor_pos)

    def set_target_velocity(self, vel: radians_per_second):
        self._motor.set(ctre.ControlMode.Velocity, vel * k_rad_per_sec_to_sensor_vel)

    def follow(self, master: _Talon):
        self._motor.follow(master._motor)

    def _set_config(self, config: Optional[TalonConfig]):
        if config is None:
            return
        if config.k_P is not None:
            self._motor.config_kP(0, config.k_P)
        if config.k_I is not None:
            self._motor.config_kI(0, config.k_I)
        if config.k_D is not None:
            self._motor.config_kD(0, config.k_D)
        if config.k_F is not None:
            self._motor.config_kF(0, config.k_F)
        if config.closed_loop_peak_output is not None:
            self._motor.configClosedLoopPeakOutput(0, config.closed_loop_peak_output)
        if config.motion_cruise_velocity is not None:
            self._motor.configMotionCruiseVelocity(config.motion_cruise_velocity * k_rad_per_sec_to_sensor_vel)
        if config.motion_acceleration is not None:
            self._motor.configMotionAcceleration(config.motion_acceleration * k_rad_per_sec_sq_to_sensor_accel)
        if config.neutral_brake is not None:
            self._motor.setNeutralMode(ctre.NeutralMode.Brake if config.neutral_brake else ctre.NeutralMode.Coast)
        if config.integral_zone is not None:
            self._motor.config_IntegralZone(0, config.integral_zone)
        if config.max_integral_accumulator is not None:
            self._motor.configMaxIntegralAccumulator(0, config.max_integral_accumulator)


class TalonFX(_Talon):
    """
    TalonFX motor controller wrapper
    """

    def init(self):
        """
        Initialize the motor controller
        """
        self._motor = ctre.TalonFX(self._can_id)
        self._set_config(self._config)
        self._motor.setInverted(self._inverted)


class TalonSRX(_Talon):
    """
    TalonSRX motor controller wrapper
    """

    def init(self):
        """
        Initialize the motor controller
        """
        self._motor = ctre.TalonSRX(self._can_id)
        self._set_config(self._config)
        self._motor.setInverted(self._inverted)


class VictorSPX(_Talon):
    """
    VictorSPX motor controller wrapper
    """

    def init(self):
        """
        Initialize the motor controller
        """
        self._motor = ctre.VictorSPX(self._can_id)
        self._set_config(self._config)
        self._motor.setInverted(self._inverted)


class TalonGroup(PIDMotor):
    """
    Group of Talon motor controllers. Used when multiple motors act as a single unit with a leader motor
    """
    motors: list[_Talon]

    def __init__(self, *motors: _Talon, config: TalonConfig = None, leader_idx: int = 0):
        super().__init__()
        self.motors = list(motors)
        for m in self.motors:
            m._config = config
        self._leader_idx = leader_idx

    def init(self):
        """
        Initialize the motor controllers
        """
        self.motors[self._leader_idx].init()
        for idx, motor in enumerate(self.motors):
            if idx != self._leader_idx:
                motor.init()
                motor.follow(self.motors[self._leader_idx])

    def set_leader_idx(self, idx: int):
        """
        Set the leader motor index (in the list of motors)

        Args:
            idx (int): Index of the leader motor
        """
        self._leader_idx = idx
        for idx, motor in enumerate(self.motors):
            if idx != self._leader_idx:
                motor.follow(self.motors[self._leader_idx])

    def get_sensor_position(self) -> radians:
        """
        Get the sensor position of the leader motor

        Returns:
            position (radians): Sensor position of the leader motor in radians
        """
        return self.motors[self._leader_idx].get_sensor_position()

    def set_sensor_position(self, pos: radians):
        """
        Set the sensor position of the leader motor
        Args:
            pos (radians): Sensor position of the leader motor in radians
        """
        self.motors[self._leader_idx].set_sensor_position(pos)

    def get_sensor_velocity(self) -> radians_per_second:
        """
        Get the sensor velocity of the leader motor
        Returns:
            velocity (radians_per_second): Sensor velocity of the leader motor in radians per second
        """
        return self.motors[self._leader_idx].get_sensor_velocity()

    def set_raw_output(self, x: float):
        """
        Set the raw output of the leader motor
        Args:
            x (float): Raw output of the leader motor
        """
        self.motors[self._leader_idx].set_raw_output(x)

    def set_target_position(self, pos: radians):
        """
        Set the target position of the leader motor
        Args:
            pos (radians): Target position of the leader motor in radians
        """
        self.motors[self._leader_idx].set_target_position(pos)

    def set_target_velocity(self, vel: radians_per_second):
        """
        Set the target velocity of the leader motor
        Args:
            vel (radians_per_second): Target velocity of the leader motor in radians per second
        """
        self.motors[self._leader_idx].set_target_velocity(vel)



from builtins import type
from dataclasses import dataclass
from typing import Optional

from rev import CANSparkMax, SparkMaxPIDController, SparkMaxRelativeEncoder, SparkMaxAlternateEncoder

from robotpy_toolkit_7407.motor import PIDMotor
from robotpy_toolkit_7407.utils.units import rev, minute, radians, radians_per_second, rad, s, rotations_per_second, \
    rotations

from robotpy_toolkit_7407.unum import Unum

from robotpy_toolkit_7407.motors.ctre_motors import hundred_ms


@dataclass
class SparkMaxConfig:
    """
    Configuration for a SparkMax motor controller

    Args:
        kP: Proportional gain
        kI: Integral gain
        kD: Derivative gain
        kF: Feedforward gain
        output_range: The minimum and maximum output of the controller as (min: float, max: float)
        idle_mode: Whether to brake or coast when the motor is not moving
    """

    k_P: Optional[float] = None
    k_I: Optional[float] = None
    k_D: Optional[float] = None
    k_F: Optional[float] = None
    output_range: Optional[tuple[float, float]] = None
    idle_mode: Optional[CANSparkMax.IdleMode] = None


rev_sensor_unit = Unum.unit("rev_sensor_u", rev / 4096, "rev sensor unit")
rev_sensor_vel_unit = rev_sensor_unit / hundred_ms
rev_sensor_accel_unit = rev_sensor_vel_unit / s

k_sensor_pos_to_radians = rev.asNumber(rad)
k_radians_to_sensor_pos = rad.asNumber(rev)
k_sensor_vel_to_rad_per_sec = (rev / minute).asNumber(rad / s)
k_rad_per_sec_to_sensor_vel = (rad / s).asNumber(rev / minute)


class SparkMax(PIDMotor):
    """
    Wrapper class for the SparkMax motor controller
    """
    _motor: CANSparkMax
    _encoder: SparkMaxRelativeEncoder
    __pid_controller: SparkMaxPIDController

    def __init__(self, can_id: int, inverted: bool = True, brushless: bool = True, config: SparkMaxConfig = None):
        """

        Args:
            can_id (int): The CAN ID of the motor controller
            inverted (bool, optional): Whether the motor is inverted. Defaults to True.
            brushless (bool, optional): Whether the motor is brushless. Defaults to True.
            config (SparkMaxConfig, None): The configuration for the motor controller. Defaults to None.
        """
        super().__init__()
        self._can_id = can_id
        self._inverted = inverted
        self._brushless = brushless
        self._config = config

    def init(self):
        """
        Initializes the motor controller, pid controller, and encoder
        """
        self._motor = CANSparkMax(
            self._can_id,
            CANSparkMax.MotorType.kBrushless if self._brushless else CANSparkMax.MotorType.kBrushed
        )
        self._motor.setInverted(self._inverted)
        self.__pid_controller = self._motor.getPIDController()
        self._encoder = self._motor.getEncoder()
        self._set_config(self._config)

    def set_raw_output(self, x: float):
        """
        Sets the raw output of the motor controller

        Args:
            x (float): The output of the motor controller (between -1 and 1)
        """
        self._motor.set(x)

    def set_target_position(self, pos: rotations):
        """
        Sets the target position of the motor controller in rotations

        Args:
            pos (float): The target position of the motor controller in rotations
        """
        self.__pid_controller.setReference(pos, CANSparkMax.ControlType.kPosition)

    def set_target_velocity(self, vel: rotations_per_second):
        """
        Sets the target velocity of the motor controller in rotations per second

        Args:
            vel (float): The target velocity of the motor controller in rotations per second
        """
        self.__pid_controller.setReference(vel, CANSparkMax.ControlType.kVelocity)

    def get_sensor_position(self) -> rotations:
        """
        Gets the sensor position of the motor controller in rotations

        Returns:
            (rotations): The sensor position of the motor controller in rotations
        """
        return self._encoder.getPosition()

    def set_sensor_position(self, pos: rotations):
        """
        Sets the sensor position of the motor controller in rotations

        Args:
            pos (rotations): The sensor position of the motor controller in rotations
        """
        self._encoder.setPosition(pos)

    def get_sensor_velocity(self) -> rotations_per_second:
        """
        Gets the sensor velocity of the motor controller in rotations per second

        Returns:
            (rotations_per_second): The sensor velocity of the motor controller in rotations per second
        """
        return self._encoder.getVelocity()

    def _set_config(self, config: SparkMaxConfig):
        if config is None:
            return
        if config.k_P is not None:
            self.__pid_controller.setP(config.k_P)
        if config.k_I is not None:
            self.__pid_controller.setI(config.k_I)
        if config.k_D is not None:
            self.__pid_controller.setD(config.k_D)
        if config.k_F is not None:
            self.__pid_controller.setFF(config.k_F)
        if config.output_range is not None:
            self.__pid_controller.setOutputRange(config.output_range[0], config.output_range[1])
        if config.idle_mode is not None:
            self._motor.setIdleMode(config.idle_mode)



from robotpy_toolkit_7407.subsystem_templates.drivetrain.differential_drivetrain import DifferentialDrivetrain
from robotpy_toolkit_7407.subsystem_templates.drivetrain.differential_drivetrain_commands import DriveArcade
from robotpy_toolkit_7407.subsystem_templates.drivetrain.swerve_drivetrain import SwerveDrivetrain, SwerveNode, SwerveGyro
from robotpy_toolkit_7407.subsystem_templates.drivetrain.swerve_drivetrain_commands import DriveSwerve



import math

from robotpy_toolkit_7407.unum import Unum
from wpimath.geometry import Rotation2d, Pose2d, Translation2d
from wpimath.kinematics import SwerveDrive4Odometry, SwerveDrive4Kinematics, SwerveModuleState, ChassisSpeeds, SwerveModulePosition
from wpimath.estimator import SwerveDrive4PoseEstimator

from robotpy_toolkit_7407.oi.joysticks import JoystickAxis
from robotpy_toolkit_7407.subsystem import Subsystem
from robotpy_toolkit_7407.utils import logger
from robotpy_toolkit_7407.utils.math import rotate_vector, bounded_angle_diff
from robotpy_toolkit_7407.utils.units import s, m, deg, rad, hour, mile, rev, meters, meters_per_second, \
    radians_per_second, radians


class SwerveNode:
    """
    Extendable class for swerve node.
    """
    motor_reversed: bool = False
    motor_sensor_offset: radians = 0

    def init(self):
        """
        Initialize the swerve node.
        """
        ...

    def set(self, vel: meters_per_second, angle_radians: radians_per_second):
        """
        Set the velocity and angle of the swerve node.

        Args:
            vel (meters_per_second): velocity of the swerve node
            angle_radians (radians_per_second): turning swerve node velocity in radians per second
        """
        self._set_angle(angle_radians, self.get_current_motor_angle() + self.motor_sensor_offset)
        self.set_motor_velocity(vel if not self.motor_reversed else -vel)

    # OVERRIDDEN FUNCTIONS
    def set_motor_angle(self, pos: radians):
        """
        Set the angle of the swerve node. Must be overridden.

        Args:
            pos (radians): angle of the swerve node in radians
        """
        ...

    def get_current_motor_angle(self) -> radians:
        """
        Get the current angle of the swerve node. Must be overridden. Must return radians.
        """
        ...

    def set_motor_velocity(self, vel: meters_per_second):
        """
        Set the velocity of the swerve node. Must be overridden.
        Args:
            vel (meters_per_second): velocity of the swerve node in meters per second
        """
        ...

    def get_motor_velocity(self) -> meters_per_second:
        """
        Get the velocity of the swerve node. Must be overridden. Must return meters per second.
        """
        ...

    # 0 degrees is facing right | "ethan is our FRC lord and saviour" - sid
    def _set_angle(self, target_angle: radians, initial_angle: radians):
        target_sensor_angle, flipped, flip_sensor_offset = SwerveNode._resolve_angles(target_angle, initial_angle)

        target_sensor_angle -= self.motor_sensor_offset

        if flipped:
            self.motor_reversed = not self.motor_reversed
            self.motor_sensor_offset += flip_sensor_offset

        self.set_motor_angle(target_sensor_angle)

    @staticmethod
    def _resolve_angles(target_angle: radians, initial_angle: radians) -> tuple[float, bool, float]:
        """
        :param target_angle: Target node angle
        :param initial_angle: Initial node sensor angle
        :return: (target_sensor_angle, flipped, flip_sensor_offset)
        """

        # Actual angle difference in radians
        diff = bounded_angle_diff(initial_angle, target_angle)

        # Should we flip
        if abs(diff) > 0.65 * math.pi:
            flip_sensor_offset = math.pi if diff > 0 else -math.pi
            diff -= flip_sensor_offset
            return diff + initial_angle, True, flip_sensor_offset

        return diff + initial_angle, False, 0


class SwerveGyro:
    """
    Extendable class for swerve gyro.
    """

    def init(self):
        """
        Initialize the swerve gyro. Overridden class.
        """
        ...

    def get_robot_heading(self) -> radians:
        """
        Get the robot heading in radians. Overridden class. Must return radians.
        """
        ...

    def reset_angle(self):
        """
        Reset the robot heading. Overridden class.
        """
        ...


class SwerveDrivetrain(Subsystem):
    """
    Swerve Drivetrain Extendable class. Contains driving functions.
    """
    n_00: SwerveNode  # Top Left
    n_01: SwerveNode  # Bottom Left
    n_10: SwerveNode  # Top Right
    n_11: SwerveNode  # Bottom Right
    gyro: SwerveGyro
    axis_dx: JoystickAxis
    axis_dy: JoystickAxis
    axis_rotation: JoystickAxis
    track_width: meters = 1
    max_vel: meters_per_second = (20 * mile / hour).asNumber(m / s)  # Maximum velocity
    max_angular_vel: radians_per_second = (4 * rev / s).asNumber(rad / s)  # Maximum angular velocity
    deadzone_velocity: meters_per_second = 0.05  # Does not run within this speed
    deadzone_angular_velocity: radians_per_second = (5 * deg / s).asNumber(rad / s)  # Will not turn within this speed
    start_pose: Pose2d = Pose2d(0, 0, 0)  # Starting pose of the robot from wpilib Pose (x, y, rotation)

    def __init__(self):
        super().__init__()
        self.kinematics: SwerveDrive4Kinematics | None = None
        self.odometry: SwerveDrive4Odometry | None = None
        self.odometry_estimator: SwerveDrive4PoseEstimator | None = None
        self.chassis_speeds: ChassisSpeeds | None = None
        self._omega: radians_per_second = 0

        self.node_translations: tuple[Translation2d] | None = None
        self.swerve_positions: tuple[SwerveModulePosition] | None = None

    def init(self):
        """
        Initialize the swerve drivetrain, kinematics, odometry, and gyro.
        """
        logger.info("initializing swerve drivetrain", "[swerve_drivetrain]")
        self.n_00.init()
        self.n_01.init()
        self.n_10.init()
        self.n_11.init()
        self.gyro.init()

        logger.info("initializing odometry", "[swerve_drivetrain]")

        self.node_translations = (
            Translation2d(-.5 * self.track_width, -.5 * self.track_width),
            Translation2d(-.5 * self.track_width, .5 * self.track_width),
            Translation2d(.5 * self.track_width, -.5 * self.track_width),
            Translation2d(.5 * self.track_width, .5 * self.track_width)
        )

        self.kinematics = SwerveDrive4Kinematics(
            *self.node_translations
        )

        self.swerve_positions = (
            SwerveModulePosition(
                distance=((self.track_width * .5)**2 + (.5 * self.track_width)**2)**.5,
                angle=Rotation2d(math.atan2(-.5 * self.track_width, -.5 * self.track_width))
            ),
            SwerveModulePosition(
                distance=((self.track_width * .5)**2 + (.5 * self.track_width)**2)**.5,
                angle=Rotation2d(math.atan2(-.5 * self.track_width, .5 * self.track_width)),
            ),
            SwerveModulePosition(
                distance=((self.track_width * .5)**2 + (.5 * self.track_width)**2)**.5,
                angle=Rotation2d(math.atan2(.5 * self.track_width, -.5 * self.track_width)),
            ),
            SwerveModulePosition(
                distance=((self.track_width * .5)**2 + (.5 * self.track_width)**2)**.5,
                angle=Rotation2d(math.atan2(.5 * self.track_width, .5 * self.track_width)),
            )
        )

        self.odometry = SwerveDrive4Odometry(
            self.kinematics,
            Rotation2d(self.gyro.get_robot_heading()),
            self.swerve_positions,
            self.start_pose
        )
        self.odometry_estimator = SwerveDrive4PoseEstimator(
            self.kinematics,
            Rotation2d(self.gyro.get_robot_heading()),
            self.swerve_positions,
            self.start_pose
        )

        logger.info("initialization complete", "[swerve_drivetrain]")

    def set_driver_centric(self, vel: (meters_per_second, meters_per_second), angular_vel: radians_per_second):
        """
        Set the driver centric velocity and angular velocity. Driver centric runs with perspective of driver.

        Args:
            vel: velocity in x and y direction as (meters per second, meters per second)
            angular_vel: angular velocity in radians per second
        """
        vel = rotate_vector(vel[0], vel[1], -self.gyro.get_robot_heading())
        self.set_robot_centric(vel, angular_vel)

    def set_robot_centric(self, vel: (meters_per_second, meters_per_second), angular_vel: radians_per_second):
        """
        Set the robot centric velocity and angular velocity. Robot centric runs with perspective of robot.
        Args:
            vel: velocity in x and y direction as (meters per second, meters per second)
            angular_vel: angular velocity in radians per second
        """
        self._omega = angular_vel  # For simulation

        if abs(vel[0]) < self.deadzone_velocity and abs(vel[1]) < self.deadzone_velocity and \
                abs(angular_vel) < self.deadzone_angular_velocity:
            self.n_00.set_motor_velocity(0)
            self.n_01.set_motor_velocity(0)
            self.n_10.set_motor_velocity(0)
            self.n_11.set_motor_velocity(0)
        else:
            self.n_00.set(*self._calculate_swerve_node(
                -.5 * self.track_width, -.5 * self.track_width,
                vel[0], vel[1], angular_vel
            ))
            self.n_01.set(*self._calculate_swerve_node(
                -.5 * self.track_width, .5 * self.track_width,
                vel[0], vel[1], angular_vel
            ))
            self.n_10.set(*self._calculate_swerve_node(
                .5 * self.track_width, -.5 * self.track_width,
                vel[0], vel[1], angular_vel
            ))
            self.n_11.set(*self._calculate_swerve_node(
                .5 * self.track_width, .5 * self.track_width,
                vel[0], vel[1], angular_vel
            ))

        module_states = (
            SwerveModuleState(
                self.n_00.get_motor_velocity(),
                Rotation2d(self.n_00.get_current_motor_angle())
            ), SwerveModuleState(
                self.n_01.get_motor_velocity(),
                Rotation2d(self.n_01.get_current_motor_angle())
            ), SwerveModuleState(
                self.n_10.get_motor_velocity(),
                Rotation2d(self.n_10.get_current_motor_angle())
            ), SwerveModuleState(
                self.n_11.get_motor_velocity(),
                Rotation2d(self.n_11.get_current_motor_angle())
            )
        )

        self.odometry.update(Rotation2d(self.gyro.get_robot_heading()), *module_states)
        self.odometry_estimator.update(Rotation2d(self.gyro.get_robot_heading()), *module_states)
        self.chassis_speeds = self.kinematics.toChassisSpeeds(module_states)

    def stop(self):
        """
        Stop the drivetrain and all pods.
        """
        self.n_00.set(0, 0)
        self.n_01.set(0, 0)
        self.n_10.set(0, 0)
        self.n_11.set(0, 0)

    @staticmethod
    def _calculate_swerve_node(node_x: meters, node_y: meters, dx: meters_per_second, dy: meters_per_second,
                               d_theta: radians_per_second) -> (meters_per_second, radians):
        tangent_x, tangent_y = -node_y, node_x
        tangent_m = math.sqrt(tangent_x ** 2 + tangent_y ** 2)
        tangent_x /= tangent_m
        tangent_y /= tangent_m

        r = math.sqrt(2) / 2
        sx = dx + r * d_theta * tangent_x
        sy = dy + r * d_theta * tangent_y

        theta = math.atan2(sy, sx)
        magnitude = math.sqrt(sx ** 2 + sy ** 2)
        return magnitude, theta



from robotpy_toolkit_7407.unum import Unum

from robotpy_toolkit_7407.motor import PIDMotor
from robotpy_toolkit_7407.oi.joysticks import JoystickAxis
from robotpy_toolkit_7407.subsystem import Subsystem
from robotpy_toolkit_7407.utils import logger
from robotpy_toolkit_7407.utils.units import rad, m, radians_per_meter, meters_per_second
class DifferentialDrivetrain(Subsystem):
    """
    Extendable differential drivetrain class.

    Args:
        m_left: leading left motor
        m_right: leading right motor
        axis_x: x-axis of the joystick
        axis_y: y-axis of the joystick
        gear_ratio: gear ratio of the drivetrain
    """
    m_left: PIDMotor = None
    m_right: PIDMotor = None
    axis_x: JoystickAxis = None
    axis_y: JoystickAxis = None
    gear_ratio: radians_per_meter

    def init(self):
        """
        Initialize the drivetrain.
        """
        logger.info("initializing differential drivetrain", "[differential_drivetrain]")
        self.m_left.init()
        self.m_right.init()
        logger.info("initialization complete", "[differential_drivetrain]")

    def set_motor_percent_output(self, left: float, right: float):
        """
        Set the percent output of the motors between -1 and 1

        Args:
            left (float): percent output of the left motor
            right (float): percent output of the right motor
        """
        self.m_left.set_raw_output(left)
        self.m_right.set_raw_output(right)

    def set_motor_velocity(self, left_vel: meters_per_second, right_vel: meters_per_second):
        """
        Set the velocity of the motors in meters per second

        Args:
            left_vel (meters_per_second): velocity of the left motor
            right_vel (meters_per_second): velocity of the right motor
        """
        self.m_left.set_target_velocity(left_vel * self.gear_ratio)
        self.m_right.set_target_velocity(right_vel * self.gear_ratio)



import math

from robotpy_toolkit_7407.unum import Unum

from robotpy_toolkit_7407.utils.units import rad


def bounded_angle_diff(theta_from: float, theta_too: float) -> float:
    """
    Finds the bounded (from -π to π) angle difference between two unbounded angles
    """
    res = math.fmod(theta_too - theta_from, 6.283185307179586)
    if res > math.pi:
        res -= 6.283185307179586
    if res < -math.pi:
        res += 6.283185307179586
    return res


def rotate_vector(x: float, y: float, theta: float) -> tuple[float, float]:
    return (
        x * math.cos(theta) - y * math.sin(theta),
        x * math.sin(theta) + y * math.cos(theta)
    )


def clamp(val: float, _min: float, _max: float):
    """
    Clamps a value between a min and max

    Args:
        val (float): value to clamp
        _min (float): min value
        _max (float): max value

    Returns:
        float: clamped value
    """
    if val < _min:
        return _min
    if val > _max:
        return _max
    return val


def ft_to_m(ft: float):
    """
    Converts feet to meters

    Args:
        ft (float): feet (float)

    Returns:
        float: meters (float)
    """
    return ft * 0.3048


def talon_sensor_units_to_inches(sensor_units: float, low_gear: bool) -> float:
    """
    Converts sensor units to inches

    Args:
        sensor_units (float): sensor units as a float
        low_gear (bool): low gear as a bool

    Returns:
        inches as a float
    """
    motor_rotations = sensor_units / 2048.0

    if low_gear:
        wheelbase_rotations = motor_rotations / 15.45  # Low gear
    else:
        wheelbase_rotations = motor_rotations / 8.21  # High gear

    inches = wheelbase_rotations * (6 * math.pi)

    return inches


def talon_sensor_units_to_meters(sensor_units: float, low_gear: bool) -> float:
    """
    Converts sensor units to meters

    Args:
        sensor_units (float):  sensor units as a float
        low_gear (bool): low gear as a bool

    Returns:
        meters as a float
    """
    return talon_sensor_units_to_inches(sensor_units, low_gear) * 0.0254


def meters_to_talon_sensor_units(meters: float, low_gear: bool) -> float:
    """
    Converts meters to sensor units

    Args:
        meters (float): meters as a float
        low_gear (float): low gear as a bool

    Returns:
        sensor units as a float
    """
    return inches_to_talon_sensor_units(meters / 0.0254, low_gear)


def inches_to_talon_sensor_units(inches: float, low_gear: bool) -> float:
    """
    Converts inches to sensor units

    Args:
        inches (float): inches as a float
        low_gear (float): low gear as a bool

    Returns:
        sensor units as a float
    """
    wheelbase_rotations = inches / (6 * math.pi)

    if low_gear:
        motor_rotations = wheelbase_rotations * 15.45  # Low gear
    else:
        motor_rotations = wheelbase_rotations * 8.21  # High gear

    sensor_units = motor_rotations * 2048.0

    return sensor_units



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['robotpy_toolkit_7407',
 'robotpy_toolkit_7407.motors',
 'robotpy_toolkit_7407.oi',
 'robotpy_toolkit_7407.pneumatics',
 'robotpy_toolkit_7407.pneumatics.pistons',
 'robotpy_toolkit_7407.sensors',
 'robotpy_toolkit_7407.sensors.color_sensors',
 'robotpy_toolkit_7407.sensors.gyro',
 'robotpy_toolkit_7407.sensors.limelight',
 'robotpy_toolkit_7407.sensors.limit_switches',
 'robotpy_toolkit_7407.sensors.odometry',
 'robotpy_toolkit_7407.sensors.photonvision',
 'robotpy_toolkit_7407.subsystem_templates',
 'robotpy_toolkit_7407.subsystem_templates.drivetrain',
 'robotpy_toolkit_7407.tests',
 'robotpy_toolkit_7407.unum',
 'robotpy_toolkit_7407.unum.units',
 'robotpy_toolkit_7407.unum.units.custom',
 'robotpy_toolkit_7407.unum.units.others',
 'robotpy_toolkit_7407.unum.units.si',
 'robotpy_toolkit_7407.utils']

package_data = \
{'': ['*']}

install_requires = \
['pynetworktables>=2021.0.0,<2022.0.0',
 'robotpy-apriltag>=2023.1.1.0,<2024.0.0.0',
 'robotpy-ctre>=2023.0.0,<2024.0.0',
 'robotpy-photonvision>=2023.1.2,<2024.0.0',
 'robotpy-rev>=2023.1.1,<2024.0.0',
 'robotpy[commands2]>=2023.1.1,<2024.0.0']

setup_kwargs = {
    'name': 'robotpy-toolkit-7407',
    'version': '0.5.49.2',
    'description': "FRC Wired Boars Team 7407's Toolkit for usage with RobotPy",
    'long_description': 'None',
    'author': 'Sidharth Rao',
    'author_email': 'sidharthmrao@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>3.9,<3.12',
}


setup(**setup_kwargs)



import math
from networktables import NetworkTables
from wpimath.geometry import Pose3d, Translation3d, Rotation3d

from robotpy_toolkit_7407.utils.units import m, deg, rad, radians
from robotpy_toolkit_7407.sensors.odometry import VisionEstimator


class Limelight:
    """
    Wrapper for the Limelight sensor.
    Connect, get, and modify limelight values and settings through the NetworkTables interface.
    """

    def __init__(self, cam_height: float, cam_angle: float, target_height: float = None, robot_ip: str = "10.74.07.2"):
        """
        Args:
            cam_height (float): Height of the limelight camera from the ground in meters.
            cam_angle (float): Camera angle from the horizontal in degrees.
            target_height (float, optional): Height of the target from the ground in meters. Defaults to camera height.
        """

        NetworkTables.initialize(server=robot_ip)
        self.table = NetworkTables.getTable("limelight")
        self.tx = 0
        self.ty = 0
        self.refs = 0
        self.k_cam_height = (cam_height * m).asNumber(m)  # Height from ground
        self.k_cam_angle: radians = (cam_angle * deg).asNumber(rad)  # Angle from horizontal
        if target_height is not None:
            self.k_h_target_height = (target_height * m).asNumber(m)
        else:
            self.k_h_target_height = self.k_cam_height

    def led_on(self):
        """Turn limelight LEDs on. Recommended to use ref_on instead.
        """
        self.table.putNumber("ledMode", 3)

    def led_off(self):
        """Turn limelight LEDs off. Recommended to use ref_off instead.
        """
        self.table.putNumber("ledMode", 1)

    def ref_on(self):
        """
        Turn on the limelight LEDs and increment the reference counter.
        """
        if self.refs == 0:
            self.led_on()
        self.refs += 1

    def ref_off(self):
        """
        Turn off the limelight LEDs and decrement the reference counter.
        """
        self.refs -= 1
        if self.refs == 0:
            self.led_off()

    def update(self):
        """Update Limelight Values to NetworkTables. Run this every loop to ensure fresh values.
        """
        c_tx = self.table.getNumber('tx', None)
        c_ty = self.table.getNumber('ty', None)
        if c_tx is None or c_ty is None:
            return "No target found."
        self.tx = c_tx
        self.ty = c_ty

    def calculate_distance(self) -> float:
        """
        Calculate distance of limelight from target. This function is not useful when the limelight and target are of
        similar height.

        Returns:
            float: Distance from target in meters
        """
        true_angle = self.k_cam_angle + self.ty
        distance = (self.k_h_target_height - self.k_cam_height) / math.tan(true_angle)
        return distance

    def get_x_offset(self) -> radians:
        """Offset on the x plane from the target in radians

        Returns:
            radians: Radian offset of the target from the center of the camera
        """
        return math.radians(self.tx)

    def get_bot_pose(self, round_to: int = None) -> list | None:
        """
        Get the robot's pose from the limelight's perspective.
        """
        bot_pose = self.table.getValue("botpose", None)
        if round_to is not None and bot_pose is not None:
            bot_pose = [round(i, round_to) for i in bot_pose]
        return bot_pose


class LimelightController(VisionEstimator):
    def __init__(self, limelight_list: list[Limelight]):
        super().__init__()
        self.limelights = limelight_list

    def get_estimated_robot_pose(self) -> list[Pose3d] | None:
        """
        Returns the robot's pose relative to the field, estimated by the limelight.
        :return: Limelight estimate of robot pose.
        :rtype: Pose3d | None
        """
        pose_list = []

        for limelight in self.limelights:
            est_pose = limelight.get_bot_pose()
            if est_pose:
                pose_list.append(Pose3d(
                    Translation3d(est_pose[0], est_pose[1], est_pose[2]),
                    Rotation3d(est_pose[3], est_pose[4], est_pose[5])
                ))
            else:
                pose_list.append(None)

        return pose_list if pose_list else None



from rev import ColorSensorV3
from wpilib import I2C
from robotpy_toolkit_7407.utils.logger import Logger


class REVColorSensor:
    """
    REVColor Sensor Wrapper for usage with I2C Multiplexer
    """

    def __init__(self,
                 sensor_port: int,
                 I2C_address: int = 0x71,
                 threshold_blue: float = 500,
                 threshold_red: float = 500,
                 threshold_green: float = 400,
                 debug: bool = False):
        """
        REVColor Sensor Wrapper for usage with I2C Multiplexer

        Args:
            sensor_port (int): Initialize with the sensor's port on multiplexer (0b0001, 0b0010, 0b0100, or 0b1000).
            I2C_address (int, optional): I2C Address of the multiplexer on RoboRio. Defaults to 0x71. Ranges from 0x70 to 0x77.
            threshold_blue (float, optional): Threshold for classification as Blue. Defaults to 500.
            threshold_red (float, optional): Threshold for classification as Red. Defaults to 500.
            threshold_green (float, optional): Used to counteract field lighting issues. Defaults to 400.
            debug (bool, optional): Use to enable debugging flags. Defaults to False.
        """

        self.port = sensor_port
        self.I2C_address = I2C_address
        self.threshold_blue = threshold_blue
        self.threshold_red = threshold_red
        self.threshold_green = threshold_green
        self.debug = debug

        self.multiplexer = I2C(I2C.Port.kMXP, self.I2C_address)
        self.sensor = ColorSensorV3(I2C.Port.kMXP)

        self.logger = Logger("ColorSensor")

    def get_val(self) -> tuple[float, float, float, int]:
        """
        Return the raw values of the color sensor's response.

        Returns:
            tuple[float, float, float, int]: R, G, B, Proximity.
        """

        self.multiplexer.writeBulk(bytes([self.port]))
        c = self.sensor.getRawColor()
        return c.red, c.green, c.blue, self.sensor.getProximity()

    def color(self) -> str:
        """Returns the color of the detected object.

        Returns:
            str: "red", "blue", or "none".
        """

        vals = self.get_val()

        if vals[0] == 0:
            if self.debug:
                self.logger.log_warning("Values not found, reinitializing color sensor...")

            self.multiplexer = I2C(I2C.Port.kMXP, self.I2C_address)
            self.sensor = ColorSensorV3(I2C.Port.kMXP)

        if vals[0] - vals[2] > self.threshold_red:
            return "red"
        elif vals[2] - vals[0] > self.threshold_blue and vals[1] > self.threshold_green:
            return "blue"

        return "none"



"""
IMPORTANT:
 - THIS DOES NOT WORK YET
 - From the perspective of the driver:
   - Positive X is forward
   - Positive Y is left
   - Positive theta is counterclockwise
"""

import math

from photonvision import PhotonUtils
from photonvision._photonvision import RobotPoseEstimator, PoseStrategy

from robotpy_toolkit_7407.sensors.photonvision.photon_target import PhotonTarget, AprilTag
from robotpy_toolkit_7407.sensors.photonvision.photon_camera import PhotonCamera
from robotpy_apriltag import AprilTagFieldLayout
from robotpy_toolkit_7407.sensors.gyro import Gyro
from wpimath.geometry import Pose3d, Translation3d, Rotation3d, Transform2d, Translation2d, Rotation2d, Transform3d


def LoadFieldLayout(json_path: str):
    return AprilTagFieldLayout(json_path)


class PhotonOdometry:
    def __init__(self, camera: PhotonCamera, field_layout: dict, gyro: Gyro,
                 start_pose=Pose3d(Translation3d(0, 0, 0), Rotation3d(roll=0, pitch=0, yaw=0))):
        self.camera = camera
        self.field_layout = self.parse_field_layout(field_layout)
        self.gyro = gyro
        self.pose_estimate = start_pose

    def refresh(self):
        self.camera.refresh()
        self.pose_estimate = self.getRobotPose()

    def getRobotPose(self, target: PhotonTarget = None):

        if target is None:
            target = self.camera.latest_best_target

        if target is None:
            return None

        try:
            field_to_target = self.field_layout.getTagPose(target.ID)
        except:
            return None

        camera_to_target = target.relative_pose
        camera_to_target_TRANSFORM = Transform2d(
            camera_to_target.translation().toTranslation2d(),
            camera_to_target.rotation().toRotation2d()
        )

        camera_to_robot = self.camera.camera_to_robot_pose
        camera_to_robot_TRANSFORM = Transform2d(
            camera_to_robot.translation().toTranslation2d(),
            camera_to_target.rotation().toRotation2d()
        )

        field_to_target = self.field_layout.getTagPose(target.ID).toPose2d()

        estimated_field_to_robot_pose = PhotonUtils.estimateFieldToRobot(
            cameraToTarget=camera_to_target_TRANSFORM,
            cameraToRobot=camera_to_robot_TRANSFORM,
            fieldToTarget=field_to_target,
        )

        estimated_field_to_robot_pose = RobotPoseEstimator(
            self.field_layout,
            PoseStrategy.CLOSEST_TO_REFERENCE_POSE,
            [
                (
                    self.camera,
                    Transform3d(
                        self.camera.camera_to_robot_pose.translation(),
                        self.camera.camera_to_robot_pose.rotation()
                    )
                )
            ]
        )

        print("ESTIMATED FIELD TO ROBOT POSE: ", estimated_field_to_robot_pose)

        gyro_angle = self.gyro.get_robot_heading()
        camera_to_target = target.relative_pose
        robot_to_camera = self.camera.camera_to_robot_pose

        theta = math.atan2(robot_to_camera.x, robot_to_camera.y) - gyro_angle
        h = (robot_to_camera.x ** 2 + robot_to_camera.y ** 2) ** .5

        field_to_camera = (
            field_to_target.x - camera_to_target.x,
            field_to_target.y - camera_to_target.y
        )

        field_to_robot = (
            field_to_camera[0] - h * math.sin(theta),
            field_to_camera[1] - h * math.cos(theta)  # Testing positive did not work
        )

        return field_to_robot

    def parse_field_layout(self, field_layout: dict):
        new_layout: AprilTagFieldLayout = None
        april_tags: list[AprilTag] = []
        length: float = None  # meters
        width: float = None  # meters

        for i in field_layout['apriltags'].keys():
            april_tags.append(
                AprilTag(
                    ID=i,
                    pose=field_layout['apriltags'][i]
                )
            )

        length = field_layout['fieldLength']
        width = field_layout['fieldWidth']
        new_layout = AprilTagFieldLayout(
            apriltags=april_tags,
            fieldLength=length,
            fieldWidth=width
        )

        return new_layout
